# Based on code taken from https://pytorch.org/docs/stable/torchvision/models.html

import torch
import torch.nn as nn

from layers.supermask_conv          import SupermaskConv2d
from layers.supermask_linear        import SupermaskLinear

__all__ = ['ResNet', 'resnet18', 'resnet34', 'resnet50', 'resnet101',
           'resnet152', 'resnext50_32x4d', 'resnext101_32x8d',
           'wide_resnet50_2', 'wide_resnet101_2']
           


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1, algo=None, sparsity=None, scale_method=None):
    """3x3 convolution with padding"""
    if algo == 'dense':
        return nn.Conv2d(
            in_planes, out_planes, kernel_size=3, stride=stride,
            padding=dilation, groups=groups, bias=False, dilation=dilation)
    else:
        return SupermaskConv2d(
            in_planes, out_planes, kernel_size=3, stride=stride,
            padding=dilation, groups=groups, bias=False, dilation=dilation, 
            algo=algo, sparsity=sparsity, scale_method=scale_method)


def conv1x1(in_planes, out_planes, stride=1, algo=None, sparsity=None, scale_method=None):
    """1x1 convolution"""
    if algo == 'dense':
        return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)
    else:
        return SupermaskConv2d(
            in_planes, out_planes, kernel_size=1, stride=stride, bias=False, 
            algo=algo, sparsity=sparsity, scale_method=scale_method)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None, use_affine_bn=False, algo=None, sparsity=None, scale_method=None):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride, algo=algo, sparsity=sparsity, scale_method=scale_method)
        self.bn1 = norm_layer(planes, affine=use_affine_bn)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes, algo=algo, sparsity=sparsity, scale_method=scale_method)
        self.bn2 = norm_layer(planes, affine=use_affine_bn)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None, use_affine_bn=False, algo=None, sparsity=None, scale_method=None):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width, algo=algo, sparsity=sparsity, scale_method=scale_method)
        self.bn1 = norm_layer(width, affine=use_affine_bn)
        self.conv2 = conv3x3(width, width, stride, groups, dilation, algo=algo, sparsity=sparsity, scale_method=scale_method)
        self.bn2 = norm_layer(width, affine=use_affine_bn)
        self.conv3 = conv1x1(width, planes * self.expansion, algo=algo, sparsity=sparsity, scale_method=scale_method)
        self.bn3 = norm_layer(planes * self.expansion, affine=use_affine_bn)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self, block, layer_list, num_classes=1000, zero_init_residual=False,
                 groups=1, width_per_group=64, replace_stride_with_dilation=None,
                 norm_layer=None, use_affine_bn=False, algo=None, sparsity=None, scale_method=None, width_factor=1):
        super(ResNet, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer

        self.inplanes = int(64 * width_factor)
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 3-element tuple, got {}".format(replace_stride_with_dilation))
        self.groups = groups
        self.base_width = width_per_group
        if algo == 'dense':
            self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False)
        else:
            self.conv1 = SupermaskConv2d(
                3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False, algo=algo, sparsity=sparsity, scale_method=scale_method)
        self.bn1 = norm_layer(self.inplanes, affine=use_affine_bn)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, int(64 * width_factor), layer_list[0], use_affine_bn=use_affine_bn, algo=algo, sparsity=sparsity, scale_method=scale_method)
        self.layer2 = self._make_layer(block, int(128 * width_factor), layer_list[1], stride=2,
                                       dilate=replace_stride_with_dilation[0],
                                       use_affine_bn=use_affine_bn, algo=algo, sparsity=sparsity, scale_method=scale_method)
        self.layer3 = self._make_layer(block, int(256 * width_factor), layer_list[2], stride=2,
                                       dilate=replace_stride_with_dilation[1],
                                       use_affine_bn=use_affine_bn, algo=algo, sparsity=sparsity, scale_method=scale_method)
        self.layer4 = self._make_layer(block, int(512 * width_factor), layer_list[3], stride=2,
                                       dilate=replace_stride_with_dilation[2],
                                       use_affine_bn=use_affine_bn, algo=algo, sparsity=sparsity, scale_method=scale_method)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        if algo == 'dense':
            self.fc = nn.Linear(int(512 * width_factor) * block.expansion, num_classes, bias=False)
        else:
            self.fc = SupermaskLinear(int(512 * width_factor) * block.expansion, num_classes, bias=False, algo=algo, sparsity=sparsity, scale_method=scale_method)

        # for m in self.modules():
        #     if isinstance(m, nn.Conv2d):
        #         nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
        #     elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
        #         nn.init.constant_(m.weight, 1)
        #         nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        # if zero_init_residual:
        #     for m in self.modules():
        #         if isinstance(m, Bottleneck):
        #             nn.init.constant_(m.bn3.weight, 0)
        #         elif isinstance(m, BasicBlock):
        #             nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False, use_affine_bn=False, algo=None, sparsity=None, scale_method=None):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride, algo=algo, sparsity=sparsity, scale_method=scale_method),
                norm_layer(planes * block.expansion, affine=use_affine_bn),
            )

        layer_list = []
        layer_list.append(block(self.inplanes, planes, stride, downsample, self.groups,
                                self.base_width, previous_dilation, norm_layer, use_affine_bn=use_affine_bn, algo=algo, sparsity=sparsity, scale_method=scale_method))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layer_list.append(block(self.inplanes, planes, groups=self.groups,
                                    base_width=self.base_width, dilation=self.dilation,
                                    norm_layer=norm_layer, use_affine_bn=use_affine_bn, algo=algo, sparsity=sparsity, scale_method=scale_method))

        return nn.Sequential(*layer_list)

    def _forward_impl(self, x):
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x

    def forward(self, x):
        return self._forward_impl(x)


def _resnet(block, layer_list, num_classes=1000, zero_init_residual=False,
            groups=1, width_per_group=64, replace_stride_with_dilation=None,
            norm_layer=None, use_affine_bn=False, algo=None, sparsity=None, scale_method=None, width_factor=None, **kwargs):
    # if width_factor != None and width_factor > 1:
    #     raise NotImplementedError
    if width_factor == None:
        width_factor = 1
    model = ResNet(
        block, 
        layer_list, 
        num_classes=num_classes, 
        zero_init_residual=zero_init_residual,
        groups=groups, 
        width_per_group=width_per_group, 
        replace_stride_with_dilation=replace_stride_with_dilation,
        norm_layer=norm_layer, 
        use_affine_bn=use_affine_bn, 
        algo=algo, 
        sparsity=sparsity, 
        scale_method=scale_method,
        width_factor=width_factor
        )
    return model


def resnet18(num_classes, **kwargs):
    r"""ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet(BasicBlock, [2, 2, 2, 2], **kwargs)


def resnet34(num_classes, **kwargs):
    r"""ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet(BasicBlock, [3, 4, 6, 3], **kwargs)


def resnet50(num_classes, **kwargs):
    r"""ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet(Bottleneck, [3, 4, 6, 3], **kwargs)


def resnet101(num_classes, **kwargs):
    r"""ResNet-101 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet(Bottleneck, [3, 4, 23, 3], **kwargs)


def resnet152(num_classes, **kwargs):
    r"""ResNet-152 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet(Bottleneck, [3, 8, 36, 3], **kwargs)


def wide_resnet50_2(num_classes, **kwargs):
    r"""Wide ResNet-50-2 model from
    `"Wide Residual Networks" <https://arxiv.org/pdf/1605.07146.pdf>`_

    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs['width_per_group'] = 64 * 2
    return _resnet(Bottleneck, [3, 4, 6, 3], **kwargs)


def wide_resnet101_2(num_classes, **kwargs):
    r"""Wide ResNet-101-2 model from
    `"Wide Residual Networks" <https://arxiv.org/pdf/1605.07146.pdf>`_

    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    kwargs['width_per_group'] = 64 * 2
    return _resnet(Bottleneck, [3, 4, 23, 3], **kwargs)
